
# === === === === === run main === === === === ===
# basic imports
import sys
import os
import cv2 as cv


# === === === === === command-line test === === === === ===

# import os, sys
# os.chdir("/Users/jameschen/Dropbox/photo_grid")
# sys.path.remove("/Users/jameschen/Dropbox/photo_grid/grid")
# from PyQt5.QtWidgets import QApplication
# import numpy as np
# import cv2
# import matplotlib.pyplot as plt
# import grid as gd

# os.chdir("/Users/jameschen/Dropbox/UAV/James/Chickpea/")
# grid = gd.GRID()
# grid.run(
#     pathImg="temp.png",
#     preset="tttttt.grid")

# ======== 20200827 segmentation ========
import os
import sys
from PyQt6.QtWidgets import QApplication
import numpy as np
import cv2
import matplotlib.pyplot as plt
import grid as gd
from .gridGUI import *
import shapefile
import numpy as np
import pandas as pd
import rasterio
import rasterio.mask

os.getcwd()

app = QApplication(sys.argv)
grid = gd.GRID()

grid.loadData(
    pathImg="/Users/poissonfish/Dropbox/photo_grid/test/map_match/demo.jpg",
    pathMap="/Users/poissonfish/Dropbox/photo_grid/test/map_match/demo.csv")
# grid.binarizeImg(k=3, lsSelect=[0, 1], valShad=0, valSmth=0)
# grid.findPlots(nRow=4, nCol=3)

g = GRID_GUI(grid, 1)  # 0:input, 1:crop, 2:kmean, 3:anchor, 4:output
app.exec_()

# ========= 20200801 multi seasons =========

# import os, sys
# sys.path
# sys.path.remove("/Users/jchen/Dropbox/projects/photo_grid/grid")
# from PyQt5.QtWidgets import QApplication
# import numpy as np
# import cv2
# import matplotlib.pyplot as plt
# os.chdir("/Users/jchen/Dropbox/projects/photo_grid/")
# import grid as gd

# import shapefile
# import numpy as np
# import pandas as pd
# import rasterio


# grid = gd.GRID()
# grid.loadData()
# pts = [[250, 500],
#        [250, 1500],
#        [750, 1500],
#        [750, 500]]
# grid.cropImg(pts)
# grid.imgs["w"]

# pts *= 1000

# imgCrop, M = gd.cropImg(img_raw, pts)

# img = grid.imgs.imgs["raw"]
# resize = 2048

# # define input coordinates
# pts = np.float32(pts)

# # assign sorted pts
# pt_NW, pt_NE, pt_SE, pt_SW = gd.sortPts(pts)

# # estimate output dimension
# img_W = (gd.euclidean(pt_NW, pt_NE) + gd.euclidean(pt_SE, pt_SW))/2
# img_H = (gd.euclidean(pt_SE, pt_NE) + gd.euclidean(pt_SW, pt_NW))/2

# # resize output dimension
# shape = find_small_shape((img_W, img_H), limit=2048)

# # generate target point
# pts2 = np.float32(
#     # NW,    NE,            SE,                   SW
#     [[0, 0], [shape[0], 0], [shape[0], shape[1]], [0, shape[1]]])

# # transformation
# H = cv2.getPerspectiveTransform(pts, pts2)
# # dst = cv2.warpPerspective(img, H, shape)
# dst = cv2.warpPerspective(np.zeros((100000, 30), dtype=np.int), H, shape)
# dst = np.array(dst).astype(np.uint8)

# plt.imshow(img)
# test = cv2.resize(img, (500, 100))
# test.shape
# plt.imshow(test)
# src[required] source/input image
# dsize[required] desired size for the output image


# plt.imshow(dst)

# rasterio.warp.aligned_target(H, shape[1], shape[0], )


# plt.imshow(grid.imgs.imgs["raw"])
# plt.imshow(grid.imgs.imgs["crop"])

# grid.cropImg(pts=[[718.3549060542798, 3950.951983298539],
#                   [5798.1503131524005, 2488.5866388308978],
#                   [7722.315240083507, 8799.847599164927],
#                   [2462.9311064718163, 10185.246346555323]])
# grid.binarizeImg(k=5, lsSelect=[0, 1], valShad=0, valSmth=0, outplot=True)

# img = grid.imgs.get("crop")
# raw = grid.imgs.get("raw")

# gimg = grid.imgs

# affine = gimg.tiff_transform
# pts_crop = gimg.pts_crop
# mat_H = gimg.mat_H

# # load agents
# agents = gimg.f_shp.shapeRecords()

# n_agents = len(agents)
# print("Test shp")
# i = 0
# # extract agent info
# pts_ag = agents[i].shape.points
# att_ag = agents[i].record

# # find the remap pts
# pts_xy = [invAffine(pts_ag[i], affine) for i in range(4)]


# # get affine old
# grid_old = gd.GRID()
# grid_old.loadData(pathImg="/Users/jameschen/Dropbox/photo_grid/test/seasons/s1.tif")

# crs1 = grid.imgs.crs
# crs2 = grid_old.imgs.crs
# crs1 == crs2

# def aff(pt, affine, affine2):
#     x = (pt[0] - affine[2]) / affine[0]
#     y = (pt[1] - affine[5]) / affine[4]
#     return (x, y)


# affine2 = grid_old.imgs.tiff_transform
# pts_xy2 = [aff(pts_ag[i], affine, affine2) for i in range(4)]

# dx = (affine[2] - affine2[2]) / affine[0]
# dy = (affine[5] - affine2[5]) / affine[4]
# dx
# dy

# plt.figure(figsize=(10, 10))
# plt.imshow(raw)
# for pt in pts_xy:
#     plt.plot(pt[0], pt[1], "x", color="red")

# plt.figure(figsize=(10, 10))
# plt.imshow(raw)
# for pt in pts_xy2:
#     plt.plot(pt[0], pt[1], "x", color="red")


# # pts_xy2 = [aff(pts_ag[i], affine2, affine2) for i in range(4)]

# # plt.figure(figsize=(10, 10))
# # plt.imshow(grid_old.imgs.get("raw"))
# # for pt in pts_xy2:
# #     plt.plot(pt[0], pt[1], "x", color="red")

# pts_crop_temp = np.matmul(
#     mat_H, np.float32(pts_xy).transpose()).transpose()

# pts_crop = np.array([pts_crop_temp[i, :2] / pts_crop_temp[i, 2]
#                     for i in range(4)], dtype=int)

# plt.figure(figsize=(10, 10))
# plt.imshow(img)
# for pt in pts_crop:
#     plt.plot(pt[0], pt[1], "x", color="red")

# agent = grid.agents.get(3, 10)
# pts_crop = [[agent.border["WEST"], agent.border["NORTH"]],
#             [agent.border["EAST"], agent.border["NORTH"]],
#             [agent.border["EAST"], agent.border["SOUTH"]],
#             [agent.border["WEST"], agent.border["SOUTH"]]]
# pts_crop = [[0, 0], [0, img.shape[0]], [img.shape[1], 0], [img.shape[1], img.shape[0]]]

# plt.imshow(img)
# for pt in pts_crop:
#     plt.plot(pt[0], pt[1], "x", color="red")


# mat_H = grid.imgs.mat_H

# # pts_rec = recover_scale(pts_crop, mat_H)
# n_points = len(pts_crop)
# # conver mat_in into 4 x 3 matrix
# mat_in = np.array([list(pts_crop[i]) + [1] for i in range(4)])

# # solve recovered matrix
# mat_recover = np.matmul(np.linalg.inv(mat_H), mat_in.transpose())

# # transpose back to the right dimension (4 x 3)
# mat_recover = mat_recover.transpose()

# # extract the first 2 elements in each point (4 x 2)
# mat_recover = [mat_recover[i, :2] / mat_recover[i, 2]
#                for i in range(n_points)]

# # return
# pts_rec = np.array(np.matrix(mat_recover)).tolist()


# plt.figure(figsize=(12, 8))
# plt.imshow(raw)
# for pt in pts_rec:
#     plt.plot(pt[0], pt[1], "x", color="red")

# tiff_transform = grid.imgs.tiff_transform
# # try remapping to Tiff coordinate
# pts_rec = [list(tiff_transform * pts_rec[i])
#            for i in range(4)]


# grid.cpuSeg(outplot=True)

# ========= 20200727 recover shapefile =========
# import os, sys
# sys.path
# # sys.path.remove("/Users/jameschen/Dropbox/photo_grid/grid")
# from PyQt5.QtWidgets import QApplication
# import numpy as np
# import cv2
# import matplotlib.pyplot as plt
# os.chdir("/Users/jameschen/Dropbox/photo_grid/")
# import grid as gd
# import shapefile
# import numpy as np
# import pandas as pd
# import rasterio
# import rasterio.mask

# os.getcwd()

# app = QApplication(sys.argv)
# grid = gd.GRID()
# os.chdir("/Users/jameschen/Dropbox/photo_grid/test/zhou")
# grid.loadData(
#     pathImg="/Users/jameschen/Dropbox/photo_grid/test/Jacob/RGB-integer.tif")

# grid.cropImg(pts=[[1266.6388308977034, 2754.9394572025053],
#                   [3689.0855949895613, 2200.78496868476],
#                   [4385.736951983298, 5430.713987473904],
#                   [1931.6242171189979, 5953.202505219207]], outplot=True)
# grid.binarizeImg(k=3, lsSelect=[0], valShad=0, valSmth=5)
# grid.findPlots(nRow=4, nCol=13, outplot=True)
# grid.cpuSeg(outplot=True)


# img = grid.imgs.get("crop")
# raw = grid.imgs.get("raw")


# agent = grid.agents.get(3, 10)
# pts_crop = [[agent.border["WEST"], agent.border["NORTH"]],
#             [agent.border["EAST"], agent.border["NORTH"]],
#             [agent.border["EAST"], agent.border["SOUTH"]],
#             [agent.border["WEST"], agent.border["SOUTH"]]]
# pts_crop = [[0, 0], [0, img.shape[0]], [img.shape[1], 0], [img.shape[1], img.shape[0]]]

# plt.imshow(img)
# for pt in pts_crop:
#     plt.plot(pt[0], pt[1], "x", color="red")


# mat_H = grid.imgs.mat_H

# # pts_rec = recover_scale(pts_crop, mat_H)
# n_points = len(pts_crop)
# # conver mat_in into 4 x 3 matrix
# mat_in = np.array([list(pts_crop[i]) + [1] for i in range(4)])

# # solve recovered matrix
# mat_recover = np.matmul(np.linalg.inv(mat_H), mat_in.transpose())

# # transpose back to the right dimension (4 x 3)
# mat_recover = mat_recover.transpose()

# # extract the first 2 elements in each point (4 x 2)
# mat_recover = [mat_recover[i, :2] / mat_recover[i, 2]
#                for i in range(n_points)]

# # return
# pts_rec = np.array(np.matrix(mat_recover)).tolist()


# plt.figure(figsize=(12, 8))
# plt.imshow(raw)
# for pt in pts_rec:
#     plt.plot(pt[0], pt[1], "x", color="red")

# tiff_transform = grid.imgs.tiff_transform
# # try remapping to Tiff coordinate
# pts_rec = [list(tiff_transform * pts_rec[i])
#            for i in range(4)]


# ========= 20200706 improve map flexibility =========
# import os, sys
# from PyQt5.QtWidgets import QApplication
# import numpy as np
# import cv2
# import matplotlib.pyplot as plt
# import grid as gd
# from .gridGUI import *
# import shapefile
# import numpy as np
# import pandas as pd
# import rasterio
# import rasterio.mask

# os.getcwd()

# app = QApplication(sys.argv)
# grid = gd.GRID()
# os.chdir("/Users/jameschen/Dropbox/photo_grid/test/map_match")

# grid.loadData(
#     pathMap="demo2.csv",
#     pathImg="demo3.png")
# grid.binarizeImg()
# grid.findPlots()
# grid.cpuSeg()

# g = GRID_GUI(grid, 4)  # 0:input, 1:crop, 2:kmean, 3:anchor, 4:output
# app.exec_()

# ========= 20200630 improve flexibility of border defining (2 rows)=========
# import os, sys
# from PyQt5.QtWidgets import QApplication
# import numpy as np
# import cv2
# import matplotlib.pyplot as plt
# import grid as gd
# from .gridGUI import *
# import shapefile
# import numpy as np
# import pandas as pd
# import rasterio
# import rasterio.mask

# os.getcwd()

# app = QApplication(sys.argv)
# grid = gd.GRID()
# os.chdir("/Users/jameschen/Dropbox/UAV/James/Chickpea/")

# grid.loadData(
#     # pathImg="70ft_20200624_5binds.tif")
#     pathImg="temp.png")
# # grid.cropImg(pts=[[704.3102766798419, 4031.569169960474],
# #                   [5925.9209486166, 2428.6561264822135],
# #                   [7674.553359683794, 8767.44861660079],
# #                   [2428.6561264822135, 10176.069169960474]])
# grid.binarizeImg(k=5, lsSelect=[0, 1], valShad=0, valSmth=5)
# grid.findPlots()
# grid.agents.setup(gmap=grid.map,
#                   gimg=grid.imgs)

# g = GRID_GUI(grid, 4)  # 0:input, 1:crop, 2:kmean, 3:anchor, 4:output
# app.exec_()

# === === === === === === Image === === === === === ===
# import os
# from PyQt5.QtWidgets import QApplication
# os.chdir("..")
# sys.path
# sys.path.remove("/Users/jameschen/Dropbox/photo_grid/grid")
# import numpy as np
# import cv2
# import matplotlib.pyplot as plt
# import grid as gd
# os.getcwd()

# path = "/Users/jameschen/Downloads/drive-download-20200928T202825Z-001/PDK52_18_09_19_AltumTIFF.tif"
# rasObj = rasterio.open(path)
# rasObj.mode

# ls_dtype = rasObj.dtypes
# nCh = rasObj.count


# npImg = np.zeros((rasObj.height, rasObj.width, 3), dtype="uint8")
# for i in range(3):
#     band = rasObj.read(i + 1)
#     if "float" in ls_dtype[i]:
#         band[band < 0] = 0
#         band_int8 = (band - band.min()) * 255 / (np.quantile(band, .999) - band.min())
#         band_int8[band_int8 > 255] = 255
#         npImg[:, :, i] = band_int8
#     else:
#         npImg[:, :, i] = band


# === === === === === === CMD DEBUG === === === === === ===
# import os
# from PyQt5.QtWidgets import QApplication
# os.chdir("..")
# sys.path
# sys.path.remove("/Users/jameschen/Dropbox/photo_grid/grid")
# import numpy as np
# import cv2
# import matplotlib.pyplot as plt
# import grid as gd
# os.getcwd()

# grid = gd.GRID()
# grid.loadData(
#     pathImg="/Users/jameschen/Downloads/drive-download-20200928T202825Z-001/PDK52_18_09_19_AltumTIFF.tif")
# plt.imshow(grid.imgs.get("raw")[:, :, :3])

# imgs = grid.imgs.get("raw")
# np.min(imgs, axis=(0, 1))
# np.max(imgs, axis=(0, 1))
# np.mean(imgs, axis=(0, 1))

# grid.cropImg(pts=[[4446.843373493975, 5406.265060240964],
#                   [3685.3975903614455, 2192.9638554216867],
#                   [1218.313253012048, 2725.975903614458],
#                   [1949.301204819277, 5984.963855421686]])

# grid.binarizeImg(k=5, lsSelect=[0, 1], features=[0, 1, 2], valShad=0, valSmth=0)
# grid.findPlots(nRow=4, nCol=13)
# grid.cpuSeg()
# grid.save(path="/Users/jameschen/Dropbox/photo_grid/test/Jacob/", prefix="jacob", h5=True)

# === === === === === === DEBUG === === === === === ===

# ### recover scale
# import pandas as pd
# import shapefile

# path = "/Users/jameschen/Dropbox/photo_grid/test/Jacob/"
# prefix = "jacob"
# h5 = True
# pathDT = os.path.join(path, prefix+"_data.csv")
# pathSp = os.path.join(path, prefix)

# # info
# imgH = grid.map.imgH
# dt = pd.read_csv(pathDT)
# mat_H = grid.imgs.mat_H
# tiff_transform = grid.imgs.tiff_transform
# pts_crop = grid.imgs.pts_crop
# n_rot = grid.imgs.n_rot
# org = (int(grid.imgs.widthRs / 2), int(grid.imgs.heightRs / 2))

# f = shapefile.Writer(pathSp)
# # define fields
# cols = dt.columns

# for col in cols:
#     instance = dt[col][0]

#     if isinstance(instance, object):
#         # characters
#         mode = "C"
#         arg1, arg2 = 20, 20
#     else:
#         # integer, floating
#         mode = "N"
#         arg1, arg2 = 10, 10

#     f.field(col, mode, arg1, arg2)

# # for idx, entry in dt.iterrows():
# entry = dt.iloc[0]
# # get agents
# row = entry["row"]
# col = entry["col"]
# agent = grid.agents.get(row, col)

# pts_crop = [[agent.border["WEST"], agent.border["NORTH"]],
#             [agent.border["EAST"], agent.border["NORTH"]],
#             [agent.border["EAST"], agent.border["SOUTH"]],
#             [agent.border["WEST"], agent.border["SOUTH"]]]

# # recover
# pts_rec = recover_scale(pts_crop, mat_H)

# # try remapping to Tiff coordinate if applicable
# try:
#     pts_rec = [list(tiff_transform * pts_rec[i]) for i in range(4)]
# except Exception as e:
#     print(e)

# # input shape file
# f.poly([pts_rec])

# # attributes
# dc = {c: entry[c] for c in dt.columns}
# f.record(**dict(dc))


###### ###### ###### ###### ###### ###### ###### ######

### repeatability
# import os
# os.chdir("/Users/jameschen/Dropbox/photo_grid")
# import grid as gd
# import pandas as pd
# import matplotlib.pyplot as plt
# from  scipy.stats import pearsonr

# grid = gd.GRID()
# pathImg = "/Users/jameschen/Dropbox/James Chen/Projects/GRID/Prototype/GRID_Demo_Croped.jpg"
# grid.loadData(pathImg=pathImg)
# grid.binarizeImg(k=3, lsSelect=[0], valShad=10, valSmth=0)
# grid.findPlots(nRow=23, nCol=12)
# grid.cpuSeg()

# gd.saveDT(grid, "/Users/jameschen/", prefix="g1_2")
# v1 = pd.read_csv("/Users/jameschen/g1_2_data.csv")["area_all"]


# grid = gd.GRID()
# pathImg = "/Users/jameschen/Dropbox/James Chen/Projects/GRID/Prototype/GRID_Demo_Croped.jpg"
# grid.loadData(pathImg=pathImg)
# grid.binarizeImg(k=3, lsSelect=[0], valShad=10, valSmth=0)
# grid.findPlots(nRow=23, nCol=12)
# grid.cpuSeg()

# gd.saveDT(grid, "/Users/jameschen/", prefix="g2_2")
# v2 = pd.read_csv("/Users/jameschen/g2_2_data.csv")["area_all"]

# pearsonr(v1, v2)


# plt.imshow(grid.imgs.get("bin"))
################


# plt.figure(figsize=(24, 24))
# # plt.imshow(grid.imgs.get("crop"))
# plt.imshow(grid.imgs.get("visSeg"))
# grid.findPlots(outplot=True)
# grid.cpuSeg()


# grid.save(pah="/Users/jameschen/", prefix="test")

# g = GRID_GUI(grid, 3)  # 0:input, 1:crop, 2:kmean, 3:anchor, 4:output
# app.exec_()


# ### open cv
# img = grid.imgs.get("crop")
# img.shape
# plt.imshow(img)

# pts1 = np.float32([[200, 200], [600, 300],
#                    [100, 800], [500, 900]])
# pts13 = np.float32([[200, 200, 1], [600, 300, 1],
#                     [100, 800, 1], [500, 900, 1]])
# fix = 123
# pts2 = np.float32([[0, 0], [fix, 0], [0, fix], [fix, fix]])
# pts23 = np.float32([[0, 0, 1], [fix, 0, 1], [0, fix, 1], [fix, fix, 1]])

# M = cv2.getPerspectiveTransform(pts1, pts2)

# pts2_f = np.matmul(M, pts13.transpose())
# pts2
# pts2_f.transpose()


# mat = img[:, :, 0]

# mat2 = img.reshape((-1, 1))
# mat2.shape

# np.multiply(M, img)
# dst = cv2.warpPerspective(img, M, (300, 300))


# plt.imshow(dst)

# return
# ========== peak searching ==========
# , prominence = (0.01, None)

# def findPeaks(img, nPeaks=0, axis=1, nSmooth=100):
#     """
#     ----------
#     Parameters
#     ----------
#     """
#     # compute 1-D signal
#     signal = img.mean(axis=(not axis)*1) # 0:nrow
#     # ignore signals from iamge frame
#     signal[:2] = 0
#     signal[-2:] = 0
#     # gaussian smooth
#     for _ in range(int(len(signal)/30)):
#         signal = np.convolve(
#             np.array([1, 2, 4, 2, 1])/10, signal, mode='same')
#     # find primary peaks
#     peaks, _ = find_peaks(signal)
#     lsDiff = np.diff(peaks)
#     medSig = statistics.median(signal[peaks])
#     stdSig = np.array(signal[peaks]).std()
#     medDiff = statistics.median(lsDiff)
#     stdDiff = np.array(lsDiff).std()
#     print(lsDiff)
#     print(medDiff)
#     print(stdDiff)
#     # get finalized peaks with distance constrain
#     peaks, _ = find_peaks(signal, distance=medDiff-stdDiff*4)
#     if nPeaks != 0:
#         if len(peaks) > nPeaks:
#             while len(peaks) > nPeaks:
#                 ls_diff = np.diff(peaks)
#                 idx_diff = np.argmin(ls_diff)
#                 idx_kick = idx_diff if (
#                     signal[peaks[idx_diff]] < signal[peaks[idx_diff+1]]) else (idx_diff+1)
#                 peaks = np.delete(peaks, idx_kick)
#         elif len(peaks) < nPeaks:
#             while len(peaks) < nPeaks:
#                 ls_diff = np.diff(peaks)
#                 idx_diff = np.argmax(ls_diff)
#                 peak_insert = (peaks[idx_diff]+peaks[idx_diff+1])/2
#                 peaks = np.sort(np.append(peaks, int(peak_insert)))
#     return peaks, signal


# pks, sig = findPeaks(grid.map.imgs[1])
# plt.plot(sig)
# plt.plot(pks, sig[pks], "x")
# plt.show()

# ========== peak searching ==========

# === === === find angles === === === ===
# import os
# os.chdir("..")
# sys.path
# sys.path.remove("/Users/jameschen/Dropbox/photo_grid/grid")
# import numpy as np
# import cv2
# import matplotlib.pyplot as plt
# import grid as gd
# os.getcwd()


# def rotateBinNdArray(img, angel):
#     # create border for the image
#     img[:, 0] = 1
#     img[0, :] = 1
#     img[:, -1] = 1
#     img[-1, :] = 1

#     # padding
#     sizePad = max(img.shape)
#     imgP = np.pad(img, [sizePad, sizePad], 'constant')

#     # rotate
#     pivot = tuple((np.array(imgP.shape[:2])/2).astype(np.int))
#     matRot = cv2.getRotationMatrix2D(pivot, angel, 1.0)
#     imgR = cv2.warpAffine(
#         imgP.astype(np.float32), matRot, imgP.shape, flags=cv2.INTER_LINEAR).astype(np.int8)

#     # crop
#     sigX = np.where(imgR.sum(axis=0) != 0)[0]
#     sigY = np.where(imgR.sum(axis=1) != 0)[0]
#     imgC = imgR[sigY[0]:sigY[-1], sigX[0]:sigX[-1]]

#     # return
#     return imgC

# # fq_sim = 10

# # signal = np.sin(2*np.pi*fq_sim* np.linspace(0, 1, 100))
# # signal = np.array([3,2,1,2,3,4,5,4,3,2,1,2,3,4,5,4,3])
# # fd = np.fft.fft(signal)/len(signal)
# # fd = fd[range(int(len(signal)/2))]

# # plt.subplot(211)
# # plt.plot(signal)
# # plt.subplot(212)
# # plt.plot(abs(fd))

# def getFourierTransform(sig):
#     sigf = abs(np.fft.fft(sig)/len(sig))
#     # sigf = abs(np.fft.fft(sig))
#     # sigf[0] = 0
#     # sigf[1] = 0
#     return sigf
#     # return sigf[2:25]


# grid = gd.GRID()
# grid.loadData(
#     "/Users/jameschen/Dropbox/James Chen/Projects/GRID/Manuscript/Remote Sensing/First Revision/demo/angle detection/ag4.png")
# grid.binarizeImg(k=3, lsSelect=[0], valShad=0, valSmth=60, outplot=False)

# img = grid.imgs.get("bin")
# row = 7
# col = 3
# degs = []
# sigs = []
# maxs = []
# plt.figure(figsize=(18, 15))
# for i in range(row):
#     deg = -i*15
#     degs.append(deg)
#     imgr = rotateBinNdArray(img, deg)
#     sigr = imgr.mean(axis=0)
#     # gaussian smooth
#     for _ in range(60):
#         sigr = np.convolve(
#             np.array([1, 2, 4, 2, 1])/10, sigr, mode='same')
#     sigrf = getFourierTransform(sigr)
#     sigabs = (sigrf)
#     plt.subplot(row, col, 1+i*col+0)
#     plt.axis('off')
#     plt.imshow(imgr)
#     plt.subplot(row, col, 1+i*col+1)
#     plt.plot(sigr)
#     plt.subplot(row, col, 1+i*col+2)
#     plt.ylim(0, 0.12)
#     plt.xticks(range(20), np.arange(1, 21))
#     plt.plot(sigabs[:20])
#     sigs.append(round(sum(sigabs), 4))
#     maxs.append(round(max(sigabs), 4))

# print(degs)
# print(maxs)
# print(sigs)

# === === === find angles === === === ===

# import os, sys
# import matplotlib.pyplot as plt
# import pandas as pd
# import numpy as np
# import h5py as h5

# f = h5.File("/Users/jameschen/Dropbox/James Chen/GRID/Manuscript/Remote Sensing/First Revision/demo/precision/GRID_case1.h5", "r")
# lsKey = np.array(list(f.keys()))
# np.random.shuffle(lsKey)

# row = 5
# col = 5
# plt.figure(0, figsize=(7, 4))
# for i in range(row):
#     for j in range(col):
#         idx = i*col + j
#         img = f[lsKey[idx]][:]
#         img[img.var(axis=(2)) == 0] = 255
#         ax = plt.subplot2grid((row, col), (i, j))
#         ax.set_yticklabels([])
#         ax.set_xticklabels([])
#         plt.tight_layout(pad=0.3)
#         plt.xticks([], [])
#         plt.yticks([], [])
#         ax.imshow(img[:, :, :3])
# plt.show()


# k = 5
# row = k
# col = 4
# i = 0

# grid.binarizeImg(k=k, lsSelect=[0], valShad=0, valSmth=0, outplot=False)

# === === === index value === === ===
# import os
# os.chdir("..")
# sys.path
# sys.path.remove("/Users/jameschen/Dropbox/photo_grid/grid")
# os.getcwd()

# import grid as gd
# import numpy as np
# import matplotlib.pyplot as plt

# grid = gd.GRID()
# # 90 deg
# pathImg = "/Users/jameschen/Dropbox/James Chen/Projects/GRID/Prototype/Alfalfa/GRID_Demo_Croped.jpg"
# grid.loadData(pathImg)
# grid.binarizeImg(k=3, lsSelect=[0], valShad=0, valSmth=0)
# grid.findPlots(nRow=23, nCol=12)
# grid.cpuSeg()

# # progress bar
# nD = grid.imgs.depth
# lsK = grid.imgs.paramKMs["lsSelect"]

# # grab info from GRID obj
# img = grid.imgs.get("crop").copy().astype(np.int)
# ch1Sub = 1 if img.shape[2] == 3 else 3  # replace NIR with Gr if it's RGB


# imgNum = (img[:, :, ch1Sub] - img[:, :, 0])
# imgDet = (img[:, :, ch1Sub] + img[:, :, 0] + 1e-8)
# imgNDVI = imgNum / imgDet

# imgDet.shape
# imgNum[1100:1200, 180:220]
# imgNum[1100:1200, :50]
# imgDet[1100:1200, 180:220]
# imgDet[1100:1200, :50]

# plt.imshow(img)
# plt.imshow(imgNum)
# plt.imshow(imgDet)
# plt.imshow(imgNDVI)

# === === === detect default rank of K === === === ===
# import grid as gd
# import numpy as np
# import matplotlib.pyplot as plt

# grid = gd.GRID()
# grid.loadData("/Users/jameschen/Dropbox/James_Git/FN/data/demo.png")
# # grid.loadData("/Users/jameschen/Dropbox/James Chen/GRID/Modeling/Rhombus.jpg")

# k = 5
# row = k
# col = 4
# i = 0

# grid.binarizeImg(k=k, lsSelect=[0], valShad=0, valSmth=0, outplot=False)

# for i in range(row):
#     imgB = (np.isin(grid.imgs.get('kmean'), i)*1).astype(np.int)
#     sigs = imgB.mean(axis=0)
#     plt.subplot(row, col, 1+i*col+0)
#     plt.imshow(imgB)
#     plt.subplot(row, col, 1+i*col+1)
#     plt.ylim(0, 1)
#     plt.plot(sigs)
#     plt.subplot(row, col, 1+i*col+2)
#     sigf = gd.getFourierTransform(sigs)
#     plt.plot(sigf)
#     plt.subplot(row, col, 1+i*col+3)
#     # scMaxF = round(max(sigf)*5, 4)  # [0, 1]
#     scMaxF = round((max(sigf)/sigf.mean())/100, 4)  # [0, 1]
#     scMean = round(1-(sigs.mean()), 4) # [0, 1]
#     scPeaks = round(len(gd.find_peaks(sigs, height=(sigs.mean()))
#                   [0])/len(gd.find_peaks(sigs)[0]), 4)
#     # scPeaks = len(gd.find_peaks(sigs)[0])
#     score = scMaxF*.25 + scMean*.25 + scPeaks*.5
#     plt.text(.3, .8, str(score), fontsize=10)
#     plt.text(.3, .6, str(scMaxF), fontsize=10)
#     plt.text(.3, .4, str(scMean), fontsize=10)
#     plt.text(.3, .2, str(scPeaks), fontsize=10)

# plt.show()

# === === === detect default rank of K === === === ===


# max(getFourierTransform(sigs))


# plt.imshow()
# plt.show()

# grid = gd.GRID()
# grid.loadData("/Users/jameschen/Dropbox/James Chen/GRID/Modeling/Rhombus.jpg")
# grid.binarizeImg(k=5, lsSelect=[0], valShad=0, valSmth=0, outplot=False)
# grid.findPlots(outplot=False)
# grid.map.dt
# # grid.cpuSeg()
# # grid.save(prefix="Pumptree2")

# app = QApplication(sys.argv)
# g = GRID_GUI(grid, 4) # 0:input, 1:crop, 2:kmean, 3:anchor, 4:output
# app.exec_()


# # ========== kmean ==========
# array = numpy.array([4, 2, 7, 1])
# temp = array.argsort()
# ranks = numpy.empty_like(temp)
# ranks[temp] = numpy.arange(len(array))

# def getRank(array):
#     sort = array.argsort()
#     rank = np.zeros(len(sort), dtype=np.int)
#     rank[sort] = np.flip(np.arange(len(array)), axis=0)
#     return rank

# test = np.array([2, 6, 87,1, 3, 6])
# getRank(test)
# temp = test.argsort()
# np.zeros(len(temp))
# rank = np.empty_like(temp)
# rank[temp] = np.flip(np.arange(len(test)), axis=0)
# np.arange(1, 7, 1)
# np.arange(6, 2, 1)

# ========== plotting ==========

# FIXME: Wrong order for the second axis

# ========== peak searching ==========
# pks, sig = gd.findPeaks(grid.map.imgs[1])
# plt.plot(sig)
# plt.plot(pks, sig[pks], "x")
# plt.show()


# grid.save(prefix="Pumptree2")

# grid = gd.GRID()
# grid.loadData("/Users/jameschen/Dropbox/James Chen/GRID/Modeling/Rhombus.jpg")
# grid.binarizeImg(k=5, lsSelect=[0], valShad=0, valSmth=0, outplot=False)
# grid.findPlots(nRow=7, nCol=6, outplot=False)
# grid.cpuSeg(outplot=True)
# grid.save(prefix="Rhombus")
# # grid.run(pathImg="/Users/jameschen/Dropbox/James Chen/GRID/Modeling/Rhombus.jpg", k=5, lsSelect=[4], path="/Users/jameschen", prefix="letmesleep")

# # with open("/Users/jameschen/Dropbox/photo_grid/Outputter_1018.grid", "wb") as file:
# #     pickle.dump(grid, file, pickle.HIGHEST_PROTOCOL)

# # grid.loadData("/Users/jameschen/Dropbox/James_Git/FN/data/demo.png")
# # grid.binarizeImg(k=3, lsSelect=[0, 1], valShad=0, valSmth=0, outplot=False)


# g = GRID_GUI(grid, 4)
# app.exec_()

#
#
#



# 3rd party imports
from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *

# self imports
from ..lib import *
from ..grid import *
from .customQt import * 


class PnCropper(QGroupBox):
    """
    """

    def __init__(self, grid):
        """
        """

        super().__init__()
        self.setStyleSheet("""
        QGroupBox::title{
            subcontrol-origin: margin;
            subcontrol-position: top center;
        }
        """)

        # attr.
        self.grid = grid
        self.layout = QVBoxLayout()
        self.wgImg = Widget_ViewCrop(grid, self.grid.imgs.get('raw'))
        self.initUI()

    def initUI(self):
        """
        """

        self.layout.addWidget(self.wgImg)
        self.setLayout(self.layout)
        self.show()

    def run(self):
        """
        """
        self.grid.cropImg(pts=self.wgImg.getPinnedPoints())


class Widget_ViewCrop(Widget_Img):

    def __init__(self, grid, img):
        super().__init__()
        self.setMouseTracking(True)
        self.grid = grid
        self.img_vis = img
        self.pos_move_prev = None
        self.pos_move = None
        self.pos_press = None
        self.pos_release = None
        self.zoom = 1
        self.ratio = 1
        self.pts = []
        self.pts_prev = [] # for rotation
        self.n_marks = 0
        self.imgH, self.imgW = img.shape[0], img.shape[1]

        # mouse event
        self.hasDrag = 0
        self.isDragDefine = 0 # is dragging to define rectangle
        self.isDragSize = 0 # is resizing?
        self.sen_resize = 30
        self.sen_rotate = 150
        self.whichState = -1
        '''
        0   4   1
           ___
        7 | 8 | 5
          |___|
        3   6   2

        9 : rotate
        '''
        #

        self.initUI()

    def initUI(self):
        super().make_rgb_img(self.img_vis)
        self.show()

    def paintEvent(self, paint_event):
        painter = QPainter(self)
        super().paintImage(painter)

        painter.setPen(QPen(Qt.red, 1, Qt.SolidLine))
        # painter.setBrush(QBrush(Qt.red, Qt.SolidPattern))
        painter.setBrush(QBrush(Qt.red, Qt.Dense7Pattern))

        if self.hasDrag:
            # has drawn points
            points = QPolygon([QPoint(self.pts[i][0], self.pts[i][1])
                              for i in range(4)])
            painter.drawPolygon(points)
        else:
            if len(self.pts) == 1:
                painter.drawLine(QPoint(self.pts[0][0], self.pts[0][1]),
                                 QPoint(self.pos_move[0], self.pos_move[1]))
            elif len(self.pts) == 2:
                points = QPolygon([QPoint(self.pts[0][0], self.pts[0][1]),
                                   QPoint(self.pts[1][0], self.pts[1][1]), 
                                   QPoint(self.pos_move[0], self.pos_move[1])])
                painter.drawPolygon(points)
            elif len(self.pts) == 3:
                points = QPolygon([QPoint(self.pts[0][0], self.pts[0][1]),
                                   QPoint(self.pts[1][0], self.pts[1][1]),
                                   QPoint(self.pts[2][0], self.pts[2][1]),
                                   QPoint(self.pos_move[0], self.pos_move[1])])
                painter.drawPolygon(points)

        painter.end()

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            # Drag to define area
            self.pos_press = (event.pos().x(), event.pos().y())
            self.pos_move_prev = self.pos_press

            if self.whichState == -1:
                # define AOI
                if len(self.pts) < 4:
                    # insert point to incomplete AOI
                    self.pts.append(self.pos_press)
                else:
                    # create new AOI
                    self.pts = [self.pos_press]
                self.hasDrag = 1 if len(self.pts) == 4 else 0
            else:
                self.isDragSize = 1
        elif event.button() == Qt.RightButton:
            # magnifier
            self.zoom = (self.zoom + 1) % 3
            self.mouseMoveEvent(event)

    def mouseMoveEvent(self, event):
        self.pos_move = (event.pos().x(), event.pos().y())

        if not self.isDragSize and self.hasDrag and len(self.pts) == 4:
            '''
            hover around when the rectangle exists
            '''

            # check distance of each corner
            dist_corner = [euclidean(self.pos_move, self.pts[i])
                           for i in range(4)]

            # check distance for each side
            len_n = euclidean(self.pts[0], self.pts[1])
            len_e = euclidean(self.pts[1], self.pts[2])
            len_s = euclidean(self.pts[2], self.pts[3])
            len_w = euclidean(self.pts[3], self.pts[0])

            dist_side_n = sum(np.array(dist_corner)[[0, 1]]) - len_n
            dist_side_e = sum(np.array(dist_corner)[[1, 2]]) - len_e
            dist_side_s = sum(np.array(dist_corner)[[2, 3]]) - len_s
            dist_side_w = sum(np.array(dist_corner)[[3, 0]]) - len_w

            dist_side = [dist_side_n, dist_side_e,
                         dist_side_s, dist_side_w]

            # concatenate the distances
            dist_all = dist_corner + dist_side

            # find the minimum distance
            idx_corner = np.argmin(dist_corner)
            idx_side = np.argmin(dist_side)
            min_all = dist_all[np.argmin(dist_all)]
            min_cor = dist_corner[idx_corner]
            min_sid = dist_side[idx_side]

            # check conditions
            if min_cor <= self.sen_resize:
                # move corner
                self.whichState = idx_corner
            elif min_sid <= self.sen_resize:
                # move side
                self.whichState = idx_side + 4
            elif cv2.pointPolygonTest(self.pts, self.pos_move, True) > 0:
                # move whole
                self.whichState = 8
            elif min_all > self.sen_resize and min_all <= self.sen_rotate:
                # roate
                self.whichState = 9
            else:
                # do nothing
                self.whichState = -1

        elif self.isDragSize:
            '''
            Draging to resize
            '''

            # capture delta
            dx = self.pos_move[0] - self.pos_move_prev[0]
            dy = self.pos_move[1] - self.pos_move_prev[1]

            # check each status
            if self.whichState <= 3:
                # update corner
                self.pts[self.whichState][0] += dx
                self.pts[self.whichState][1] += dy
            elif self.whichState > 3 and self.whichState < 8:
                if self.whichState == 4:
                    # update side N
                    idx1, idx2 = 0, 1
                elif self.whichState == 5:
                    # update side E
                    idx1, idx2 = 1, 2
                elif self.whichState == 6:
                    # update side S
                    idx1, idx2 = 2, 3
                elif self.whichState == 7:
                    # update side W
                    idx1, idx2 = 3, 0
                for i in [idx1, idx2]:
                    self.pts[i][0] += dx
                    self.pts[i][1] += dy
            elif self.whichState == 8:
                # update whole
                for i in range(4):
                    self.pts[i][0] += dx
                    self.pts[i][1] += dy
            elif self.whichState == 9:
                # rotate
                org = np.mean(self.pts_prev, axis=0)
                v_prev = self.pos_press - org
                v_cur = self.pos_move - org
                angle = find_angle(v_prev, v_cur)
                self.pts = np.array(
                    rotatePts(self.pts_prev, angle, org), dtype=int).copy()

        # cursor
        magArea = int(min(self.rgX[1] - self.rgX[0], self.rgY[1] - self.rgY[0]) / 5)
        if self.whichState < 8:
            if self.zoom != 0:
                magnifying_glass(self, event.pos(),
                                area=magArea, zoom=self.zoom * 2.5)
            else:
                self.setCursor(QCursor(Qt.ClosedHandCursor))
        elif self.whichState == 8:
            self.setCursor(QCursor(Qt.SizeAllCursor))
        elif self.whichState == 9:
            size = 32
            pixmap = QPixmap(os.path.join(
                self.grid.user.dirGrid, "res/rotate.png")).scaled(size, size)
            self.setCursor(QCursor(pixmap))

        self.pos_move_prev = self.pos_move
        self.repaint()

    def mouseReleaseEvent(self, event):
        # pt_mouse = event.pos()

        # update points order while moving
        if self.hasDrag:
            self.pts = sortPts(self.pts)
            self.pts_prev = self.pts.copy()

        self.resetStatus()

    def getPinnedPoints(self):
        self.ratio = (self.imgW) / (self.width()) if self.isFitWidth else (self.imgH) / (self.height())
        pts = [[(pt[0] - self.rgX[0]) * (self.ratio),
                (pt[1] - self.rgY[0]) * (self.ratio)] for pt in self.pts]
        if len(pts) < 4:
            pts = [[0, 0], [self.imgW, 0],
                   [self.imgW, self.imgH], [0, self.imgH]]
        return pts

    def resetStatus(self):
        self.isDragDefine = 0
        self.isDragSize = 0


# def getPerpDist(pt1, pt2, pt0):
#     """
#     Find the perpendicular distance from a point (pt0) to aline (pt1, pt2)
#     """
#     slp = (pt1[1] - pt2[1]) / (pt1[0] - pt2[0])

#     if not np.isinf(slp):
#         itc = pt1[1] - (slp * pt1[0])
#         # ax + by + c = 0
#         a = slp
#         b = -1
#         c = itc

#         # numerator
#         num = abs(a * pt0[0] + b * pt0[1] + c)
#         # denominator
#         den = np.sqrt(a ** 2 + b ** 2)

#         # return
#         return num / den
#     else:
#         # if is a vertical line
#         return abs(pt0[0] - pt1[0])


# def getShiftPts(pt1, pt2, pt0):
#     """
#     Find the updated point (pt1*, pt2*) from a new line defined by point (pt0)
#     eq: ax + by = c
#     """
#     slp = (pt1[1] - pt2[1]) / (pt1[0] - pt2[0])
    
#     if slp == 0:
#         # horizon line
#         return [pt1[0], pt0[1]], [pt2[0], pt0[1]]
#     elif np.isinf(slp):
#         # vertical line
#         return [pt0[0], pt1[1]], [pt0[0], pt2[1]]

#     else:
#         itc = pt0[1] - slp * pt0[0]
#         eq = (slp, -1, -itc)  # ax + by = c

#         # find the perpendicular slope
#         slp_perp = - 1 / slp

#         # find the interaction
#         # solve via
#         # a1x + b1y = c1
#         # a2x + b2y = c2
#         itc_1 = pt1[1] - slp_perp * pt1[0]
#         eq_1 = (slp_perp, -1, -itc_1)
#         pt1_new = np.linalg.solve(np.array([[eq[0], eq[1]],  # a
#                                             [eq_1[0], eq_1[1]]]),  # b
#                                 np.array([eq[2], eq_1[2]]))  # c

#         itc_2 = pt2[1] - slp_perp * pt2[0]
#         eq_2 = (slp_perp, -1, -itc_2)
#         pt2_new = np.linalg.solve(np.array([[eq[0], eq[1]],  # a
#                                             [eq_2[0], eq_2[1]]]),  # b
#                                 np.array([eq[2], eq_2[2]]))  # c

#         # return
#         return pt1_new, pt2_new

# getShiftPts([0, 6], [4, 9], [0, 9])
# getShiftPts([1, 3], [5, 3], [4, 6])
# np.linalg.solve(np.array([[3, 2], [5, 2]]),
#                 np.array([1, 3]))

        



# basic imports
import numpy as np
import sys

# 3-rd party imports
from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *

# self import
from ..lib import *


class Widget_Img(QWidget):
    '''
    Will keep imgRaw, imgVis and imgQmap
    '''

    def __init__(self):
        super().__init__()
        '''attr'''
        # self.img_raw = img
        # self.img_vis = img[:, :, :3].copy()
        self.qimg = None
        self.isFitWidth = None
        self.rgX, self.rgY = (0, 0), (0, 0)
        self.sizeImg = (0, 0)

    def make_rgb_img(self, img):
        self.qimg = getRGBQImg(img[:, :, :3])
        self.updateDim()

    def make_bin_img(self, img):
        self.qimg = getBinQImg(img)
        self.updateDim()

    def make_idx8_img(self, img, k):
        self.qimg = getIdx8QImg(img, k)
        self.updateDim()

    def make_gray_img(self, img):
        self.qimg = getGrayQImg(img)
        self.updateDim()

    def updateDim(self):
        self.sizeImg = self.qimg.size().scaled(self.rect().size(),
                                               Qt.KeepAspectRatio)
        if self.sizeImg.width() == self.width():
            self.isFitWidth = True
            marginY = int((self.height()-self.sizeImg.height())/2)
            self.rgX = (0, self.sizeImg.width())
            self.rgY = (marginY, marginY+self.sizeImg.height())
        elif self.sizeImg.height() == self.height():
            self.isFitWidth = False
            marginX = int((self.width()-self.sizeImg.width())/2)
            self.rgX = (marginX, marginX+self.sizeImg.width())
            self.rgY = (0, self.sizeImg.height())

    def isInRange(self, x, y):
        if x >= self.rgX[0] and x <= self.rgX[1] and\
           y >= self.rgY[0] and y <= self.rgY[1]:
            return True
        else:
            return False

    def getImgRange(self):
        return self.rgX, self.rgY

    def paintImage(self, painter):
        painter.setRenderHint(QPainter.Antialiasing, True)
        self.updateDim()
        painter.drawPixmap(self.rgX[0], self.rgY[0], self.sizeImg.width(),
                           self.sizeImg.height(), self.qimg)


def getRGBQImg(img):
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w*3, QImage.Format_RGB888)
    return QPixmap(qImg)


def getBinQImg(img):
     h, w = img.shape[0], img.shape[1]
     qImg = QImage(img.astype(np.uint8).copy(), w, h, w*1, QImage.Format_Indexed8)
     qImg.setColor(0, qRgb(0, 0, 0))
     qImg.setColor(1, qRgb(241, 225, 29))
     return QPixmap(qImg)


def getIdx8QImg(img, k):
    colormap = [qRgb(228, 26, 28),
                qRgb(55, 126, 184),
                qRgb(77, 175, 74),
                qRgb(152, 78, 163),
                qRgb(255, 127, 0),
                qRgb(255, 255, 51),
                qRgb(166, 86, 40),
                qRgb(247, 129, 191),
                qRgb(153, 153, 153)]
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w*1, QImage.Format_Indexed8)
    for i in range(k):
        qImg.setColor(i, colormap[i])
    return QPixmap(qImg)


def getGrayQImg(img):
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w*1, QImage.Format_Grayscale8)
    return QPixmap(qImg)


def magnifying_glass(widget, pos, area=200, zoom=4):
    size = int(area/zoom)
    pixmap = widget.grab(
        QRect(QPoint(pos.x()-int(size/2), pos.y()-int(size/2)), QSize(size, size)))
    try:
        rate_screen = size / pixmap.width()
        pixmap = pixmap.scaled(int(area/rate_screen), int(area/rate_screen))
        painter = QPainter(pixmap)
        'Rect'
        pen = QPen()
        pen.setWidth(2)
        pen.setColor(Qt.black)
        painter.setPen(pen)
        # define rect
        rect = QRect(QPoint(0, 0), pixmap.size()*rate_screen)
        # draw rect
        painter.drawRect(rect)
        '''Cursor'''
        pen.setWidth(2)
        pen.setColor(Qt.red)
        painter.setPen(pen)
        size_m = 10
        space = 4
        # define lines
        line1 = QLine(QPoint(size_m, 0), QPoint(space, 0))
        line2 = QLine(QPoint(0, size_m), QPoint(0, space))
        line3 = QLine(QPoint(0, -space), QPoint(0, -size_m))
        line4 = QLine(QPoint(-size_m, 0), QPoint(-space, 0))
        line1.translate(pixmap.rect().center() * rate_screen - QPoint(0, 0))
        line2.translate(pixmap.rect().center() * rate_screen - QPoint(0, 0))
        line3.translate(pixmap.rect().center() * rate_screen - QPoint(0, 0))
        line4.translate(pixmap.rect().center() * rate_screen - QPoint(0, 0))
        # draw lines
        painter.drawLine(line1)
        painter.drawLine(line2)
        painter.drawLine(line3)
        painter.drawLine(line4)

        # # remove central
        # pt_center = pixmap.rect().center() * rate_screen
        # x, y = pt_center.x(), pt_center.y()
        # painter.eraseRect(QRect(x-2, y-2, 4, 4))
        '''finish'''
        painter.end()
        cursor = QCursor(pixmap)
        widget.setCursor(cursor)
    except Exception:
        '''not in a valid region'''


def drawCross(x, y, painter, size=2):
    l1_st_x, l1_st_y = x-size, y-size
    l1_ed_x, l1_ed_y = x+size, y+size
    l2_st_x, l2_st_y = x-size, y+size
    l2_ed_x, l2_ed_y = x+size, y-size
    painter.drawLine(int(l1_st_x), int(l1_st_y), int(l1_ed_x), int(l1_ed_y))
    painter.drawLine(int(l2_st_x), int(l2_st_y), int(l2_ed_x), int(l2_ed_y))


def drawTriangle(x, y, dir, painter, range=7, peak=30):
    path = QPainterPath()
    path.moveTo(x, y)
    if dir == 'North':
        path.lineTo(x-range, y+peak)
        path.lineTo(x+range, y+peak)
    elif dir == 'South':
        path.lineTo(x-range, y-peak)
        path.lineTo(x+range, y-peak)
    elif dir == 'West':
        path.lineTo(x+peak, y-range)
        path.lineTo(x+peak, y+range)
    elif dir == 'East':
        path.lineTo(x-peak, y-range)
        path.lineTo(x-peak, y+range)
    path.lineTo(x, y)
    painter.drawPath(path)



__author__ = "Chun-Peng James Chen"
__version__ = "1.2.23"
__update__ = "Aug 5, 2022"

# imports
import subprocess
import json
import sys
from urllib import request
from pkg_resources import parse_version

if "__main__" not in sys.argv[0]:
    # prevent from re-show welcome message in gridGUI
    # welcome message
    print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
    print("                 Welcome to GRID Ver.%s    " % __version__)
    print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
    print("Author      : James Chen <niche@vt.edu>    ")
    print("Last update : %s              " % __update__)
    print("User manual : https://poissonfish.github.io/GRID/")

    if "-m" not in sys.argv[0]:
        # if in the command-line environment
        print("    Try 'python -m grid' in Terminel to launch GRID GUI,")
        print("         as command-line version is not ready yet.")
    print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
    print("Recent update ")
    print("    - Upgrade the GUI framework from PyQt5 to PyQt6")
    print("    - Support images with huge dimensions (> 32767)")
    print("    - Add CRS to shapefiles (.prj) ")
    print("    - Support ESRI shapefile compatible in QGIS    ")
    print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")

# self update
try:
    url = 'https://pypi.python.org/pypi/photo_grid/json'
    releases = json.loads(request.urlopen(url).read())['releases']
    new_version = sorted(releases, key=parse_version, reverse=True)[0]
    if __version__ != new_version:
        # Dialog
        ans = None
        bol_ans = None
        possible_pos_ans = ["y", "Y", "yes"]
        possible_neg_ans = ["n", "N", "no"]

        while bol_ans is None:
            ans = input(
                "A newer version of GRID (ver. %s) is now available, upgrade? (y/n) " % new_version)
            if ans in possible_pos_ans:
                bol_ans = True
            elif ans in possible_neg_ans:
                bol_ans = False

        if bol_ans:
            subprocess.check_call([sys.executable,
                                   '-m', 'pip', 'install',
                                   'photo_grid==%s' % new_version, '--upgrade'])
            print("\n")
            print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
            print("          Please re-launch GRID to finish the update")
            print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
            print("\n")
            quit()
except Exception:
    print("\n")
    print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
    print("     Sorry, we currently have issue updating your GRID.")
    print("~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~*~~~~~~~~~")
    print("\n")


# self imports
from .grid import *


# basic imports
import numpy as np

# 3rd party imports
from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *

# self imports
from .customQt import *

class PnAnchor(QWidget):
    """
    """

    def __init__(self, grid):
        super().__init__()
        self.grid = grid
        # compute
        self.grid.findPlots()
        self.setFocusPolicy(Qt.StrongFocus)
        self.setFocus()

        # major
        self.layout = QHBoxLayout()
        self.grRight = QWidget()
        self.wgImg = WidgetAnchor(grid)

        self.recImg = QRect(0, 0, 0, 0)
        self.recAxs = QRect(0, 0, 0, 0)
        self.recRight = QRect(0, 0, 0, 0)

        # Right Panel
        self.loRight = QVBoxLayout()
        self.sc_right = QScrollArea()
        self.sc_right.setStyleSheet("QScrollBar {width:0px;}")
        self.sc_right.setWidgetResizable(True)

        # axes
        self.grAxis = [QGroupBox("Major axis"), QGroupBox("Minor axis: ")]
        self.loAxis = [QGridLayout(), QGridLayout()]
        self.lbAg = [QLabel("Angle"), QLabel("Angle")]
        self.rbMajAg = [QRadioButton("0°"), QRadioButton("90°")]
        self.dlMinAg = QDial()
        self.lbTk = [QLabel("# of ticks"), QLabel("# of ticks")]
        self.spbTk = [QSpinBox(), QSpinBox()]
        for spb in self.spbTk:
            spb.setMaximum(1000)

        self.mtp = 1  # for slider

        # Tools
        self.idx_tool = 0 # 0 for major, 1 for minor
        self.gr_tool = QGroupBox("Adjust centeroids")
        self.lo_tool = QHBoxLayout()
        self.rb_maj = QRadioButton("Major axis")
        self.rb_min = QRadioButton("Minor axis")

        # Display
        self.gr_dis = QGroupBox("Display")
        self.lo_dis = QHBoxLayout()
        self.rb_bin = QRadioButton("Binary (A)")
        self.rb_rgb = QRadioButton("RGB (S)")

        # reset
        self.btReset = QPushButton("Reset")

        # mouse event
        self.idxAnc = -1 # which tick
        self.idxMaj = 0 # which majar angle (0 or 90)
        self.ptX = -1
        self.ptY = -1
        self.ptXpress = -1
        self.ptYpress = -1
        self.new_itc = -1

        # UI
        self.switch = True
        self.initUI()

        # show
        self.show()
        self.wgImg.updateDim()

    def initUI(self):
        agMaj, agMin = self.grid.map.angles

        # RIGHT: major axis
        if agMaj == 0:
            self.idxMaj = 0
            self.rbMajAg[0].setChecked(True)
            self.wgImg.isMajV = True
        else:
            self.idxMaj = 1
            self.rbMajAg[1].setChecked(True)
            self.wgImg.isMajV = False

        self.spbTk[0].setValue(len(self.grid.map.itcs[0]))
        self.loAxis[0].addWidget(self.lbAg[0], 0, 0)
        self.loAxis[0].addWidget(self.rbMajAg[0], 0, 1)
        self.loAxis[0].addWidget(self.rbMajAg[1], 0, 2)
        self.loAxis[0].addWidget(self.lbTk[0], 1, 0)
        self.loAxis[0].addWidget(self.spbTk[0], 1, 1, 1, 2)
        self.grAxis[0].setLayout(self.loAxis[0])
        self.loRight.addWidget(self.grAxis[0])

        # RIGHT: minor axis
        self.grAxis[1].setTitle("Minor axis")
        self.lbAg[1].setText("Angle: %d" % agMin)
        self.dlMinAg.setRange(-int(90/self.mtp), int(90/self.mtp))
        self.dlMinAg.setValue(int(agMin/self.mtp))
        self.dlMinAg.setPageStep(3)
        self.dlMinAg.setNotchesVisible(True)
        self.dlMinAg.setNotchTarget(5)
        self.spbTk[1].setValue(len(self.grid.map.itcs[1]))
        self.loAxis[1].addWidget(self.lbAg[1], 0, 0, 1, 1)
        self.loAxis[1].addWidget(self.dlMinAg, 0, 1, 1, 2)
        self.loAxis[1].addWidget(self.lbTk[1], 1, 0, 1, 1)
        self.loAxis[1].addWidget(self.spbTk[1], 1, 1, 1, 2)
        self.grAxis[1].setLayout(self.loAxis[1])
        self.loRight.addWidget(self.grAxis[1])

        # RIGHT: functions
        self.rbMajAg[0].toggled.connect(lambda: self.changeAngle(idx=0))
        self.rbMajAg[1].toggled.connect(lambda: self.changeAngle(idx=0))
        self.dlMinAg.valueChanged.connect(lambda: self.changeAngle(idx=1))
        self.spbTk[0].valueChanged.connect(lambda: self.changePeaks(idx=0))
        self.spbTk[1].valueChanged.connect(lambda: self.changePeaks(idx=1))
        self.rb_maj.toggled.connect(self.switchTool)
        self.rb_min.toggled.connect(self.switchTool)
        self.rb_bin.toggled.connect(self.displayImage)
        self.rb_rgb.toggled.connect(self.displayImage)

        # RIGHT: tool
        self.rb_maj.setChecked(True)
        self.lo_tool.addWidget(self.rb_maj)
        self.lo_tool.addWidget(self.rb_min)
        self.gr_tool.setLayout(self.lo_tool)
        self.loRight.addWidget(self.gr_tool)

        # RIGHT: display
        self.rb_bin.setChecked(True)
        self.lo_dis.addWidget(self.rb_bin)
        self.lo_dis.addWidget(self.rb_rgb)
        self.gr_dis.setLayout(self.lo_dis)
        self.loRight.addWidget(self.gr_dis)

        # RIGHT: comp
        self.loRight.addWidget(self.btReset)
        self.grRight.setLayout(self.loRight)
        self.sc_right.setWidget(self.grRight)

        # LEFT IMG: mouse tracking
        # self.wgImg.setMouseTracking(True)

        # Main
        policyRight = QSizePolicy(QSizePolicy.Policy.Preferred,
                                  QSizePolicy.Policy.Preferred)
        policyRight.setHorizontalStretch(2)
        self.sc_right.setSizePolicy(policyRight)

        policyLeft = QSizePolicy(QSizePolicy.Policy.Preferred,
                                 QSizePolicy.Policy.Preferred)
        policyLeft.setHorizontalStretch(3)
        self.wgImg.setSizePolicy(policyLeft)

        self.layout.addWidget(self.wgImg)
        self.layout.addWidget(self.sc_right)

        self.setLayout(self.layout)

    def paintEvent(self, event):
        try:
            self.updatePlots()
        except Exception as e:
            print(e)

        # '''rect'''
        # pen.setWidth(1)
        # pen.setColor(Qt.black)
        # painter.setPen(pen)
        # painter.setBrush(Qt.transparent)
        # painter.drawRect(self.rec_acr_c)
        # painter.drawRect(self.rec_acr_r)

    def updatePlots(self):
        self.updateMajorLines()
        self.updateAgents()

    def changeAngle(self, idx):
        # current angle
        if idx == 0:
            # if major axis
            angle = 0 if self.rbMajAg[0].isChecked() else 90
            # oposite angle
            angleOp = self.dlMinAg.value() * self.mtp
        else:
            # if minor axis
            angle = self.dlMinAg.value() * self.mtp
            # oposite angle
            angleOp = 0 if self.rbMajAg[0].isChecked() else 90
            # change title
            self.lbAg[1].setText("Angle: %d°" % angle)

        # major angle
        self.wgImg.isMajV = self.rbMajAg[0].isChecked()
        self.idxMaj = 0 if self.wgImg.isMajV else 1  # v -> 0°

        # print("before")
        # print("ops:%2f" % (angleOp))
        # print("self:%2f" % (angle))
        # print("after")
        # angle difference between two axes
        # if difference is greater than 0
            # # if difference is greater than 90 degrees
            # if angleDiff > 90:
            #     print("greater than 90")
            #     if angle > 0:
            #         value = (angle - 90) / self.mtp
            #     else:
            #         value = (angle + 90) / self.mtp
            #     # if idx == 0:
            #     #     value = min(angle + 90, 90) / self.mtp
            #     # else:
            #     #     value = max(-90, angle - 90) / self.mtp
            #     print("ops:%2f" % (value * self.mtp))
            #     print("self:%2f" % (angle))
            #     self.dlAg[1 - idx].setValue(int(value))
            # # if different side
            # elif angle * angleOp < 0:
            #     print("less than 90, different sides")
            #     # force the current one equal to 0
            #     angleOp = (angle - 90) if angle > 0 else (angle + 90)
            #     self.dlAg[1 - idx].setValue(int(angleOp / self.mtp))
            # # same side
            # else:
            #     print("less than 90, same sides")
            #     # current is +-90, push opp
            #     if abs(angle) > abs(angleOp):
            #         angleOp = (angle - 90) if angle > 0 else (angle + 90)
            #         self.dlAg[1 - idx].setValue(int(angleOp / self.mtp))
            #     # opp is +-90
            #     else:
            #         angleOp = 90 if angleOp > 0 else -90
            #         self.dlAg[1 - idx].setValue(int(angleOp / self.mtp))

            # if difference is less than 90 degrees and neither of them is 0
            # elif angle * angleOp > 0:
            #     # force the current one equal to 0
            #     angle = 0
            #     self.dlAg[idx].setValue(0)
        self.grid.updateCenters(idx, angle=angle)
        self.switch = False
        self.spbTk[0].setValue(len(self.grid.map.itcs[0]))
        self.spbTk[1].setValue(len(self.grid.map.itcs[1]))
        self.switch = True
        self.displayImage()
        self.repaint()

    def changePeaks(self, idx):
        if self.switch:
            nPeaks = self.spbTk[idx].value()
            self.grid.updateCenters(idx, nPeaks=nPeaks)
            self.displayImage()
            self.repaint()

    # def switchAngle(self, idx):
    #     self.idxAx = idx
    #     self.displayImage()
    #     self.grAxis[idx].setChecked(True)
    #     self.grAxis[int(not idx)].setChecked(False)
    #     self.repaint()

    def switchTool(self):
        self.idx_tool = 0 if self.rb_maj.isChecked() else 1
        self.wgImg.isMajTool = self.rb_maj.isChecked()
        self.wgImg.repaint()

    def displayImage(self):
        self.idxAx = 0 ###NOTE
        if self.rb_bin.isChecked():
            self.wgImg.make_bin_img(self.grid.map.imgBin)
        else:
            self.wgImg.make_rgb_img(self.grid.map.imgRGB)
        self.wgImg.repaint()

    def updateMajorLines(self):
        ptsRaw = self.grid.map.sigs[0]
        if self.idxMaj == 0: # which major angle (0° or 90°)
            rgSrc = (0, self.grid.map.imgW)
            rgDst = self.wgImg.rgX
        else:
            rgSrc = (0, self.grid.map.imgH)
            rgDst = self.wgImg.rgY

        pts = rescale(ptsRaw, rgSrc, rgDst)
        self.wgImg.ptMajLine = pts

    def updateAgents(self):
        # fetch info
        gmap = self.grid.map
        agCr = gmap.angles[0]
        agOp = gmap.angles[1]
        agDiff = agOp
        agAbs = abs(agDiff)

        imgH, imgW = gmap.imgH, gmap.imgW
        qimgH, qimgW = self.wgImg.sizeImg.height(), self.wgImg.sizeImg.width()
        ratio = sum([qimgW / imgW, qimgH / imgH]) / 2

        # current axis
        # self.wgImg.ptMajLine = self.wgAxs.pts

        # another axis
        self.wgImg.slp = 1 / np.tan(np.pi / 180 * agDiff)
        sigs = gmap.sigs[1]
        itc = getCardIntercept(sigs, agDiff, imgH)

        # if agCr % 90 == 0:
        #     # bugmsg("case 1")
        # elif agOp % 90 == 0:
        #     itc = np.cos(np.pi / 180 * agAbs) * gmap.imgHr[1] + \
        #             sigs / np.sin(np.pi / 180 * agAbs)
        #     if agDiff < 0:
        #         # bugmsg("case 2")
        #         None
        #     else:
        #         # bugmsg("case 3")
        #         itc = gmap.imgHr[idxCr] - itc
        # else:
        #     seg1 = gmap.imgH * np.cos(np.pi / 180 * abs(agCr))
        #     seg2 = (sigs - gmap.imgH * np.sin(np.pi / 180 * abs(agOp))) / \
        #         np.sin(np.pi / 180 * (abs(agOp) + abs(agCr)))
        #     itc = seg1 + seg2
        #     if agCr > 0:
        #         # bugmsg("case 4")
        #         None
        #     else:
        #         # bugmsg("case 5")
        #         itc = gmap.imgHr[idxCr] - itc

        self.wgImg.itcs = itc * ratio

        # slp = -1/np.tan(np.pi/180*agAbs) if agAbs < 0 else 1/np.tan(np.pi/180*agAbs)
        # sigs = self.grid.map.sigs[idxOp]
        # if idxOp==1:
        #     self.wgImg.slp = slp
        #     X = (sigs/np.sin(np.pi/180*agAbs)) + \
        #         np.cos(np.pi/180*agAbs)*self.grid.map.imgHr[idxOp]
        #     self.wgImg.itcs = (qimgH - X*ratio)
        # else:
        #     self.wgImg.slp = -slp
        #     segA = sigs/np.sin(np.pi/180*agAbs)
        #     segB = np.sin(np.pi/180*agAbs)*self.grid.map.imgWr[idxOp]
        #     self.wgImg.itcs = segA*ratio + (qimgW - segB*ratio)

    def updateDim(self):
        self.recImg = self.wgImg.geometry()
        self.recRight = self.grRight.geometry()

    def getPtGui2Map(self, pt, axis=0):
        rgWg = self.wgImg.getImgRange()[axis]
        if axis == 0:
            rgMap = (0, self.grid.map.imgW-1)
            value = rescale(values=pt-self.recImg.x(),
                            scaleSrc=rgWg, scaleDst=rgMap)
        else:
            rgMap = (0, self.grid.map.imgH-1)
            value = rescale(values=pt-self.recImg.y(),
                            scaleSrc=rgWg, scaleDst=rgMap)
        return value

    def mousePressEvent(self, event):
        # ptX, ptY are coordinate in map
        pos = event.pos()
        objMap = self.grid.map
        self.updateDim()
        if self.recImg.contains(pos):
            self.ptX = self.getPtGui2Map(pos.x(), axis=0)
            self.ptY = self.getPtGui2Map(pos.y(), axis=1)
            self.ptXpress, self.ptYpress = self.ptX, self.ptY
            if self.idx_tool == 0:
                # major axis  0° or 90°
                ptPress = self.ptX if self.idxMaj == 0 else self.ptY
                # search index by signals
                self.idxAnc = np.abs(
                    ptPress-objMap.sigs[0]).argmin()
            else:
                # minor axis
                slope = objMap.slps[1]
                intercepts = objMap.itcs[1]
                # search index by intercepts, will be fine for angle < 0
                if objMap.angles[1] == 0:
                    self.idxAnc = np.abs(self.ptX-objMap.sigs[1]).argmin()
                else:
                    ls_y = np.array(intercepts) + self.ptX * slope
                    self.idxAnc = np.abs(self.ptY-ls_y).argmin()

    def mouseMoveEvent(self, event):
        pos = event.pos()
        objMap = self.grid.map
        if self.idxAnc != -1:
            if self.idx_tool == 0:
                # major axis
                ptPos = pos.x() if self.idxMaj == 0 else pos.y()
                rg_wgImg = self.wgImg.rgX if self.idxMaj == 0 else self.wgImg.rgY
                size_img = objMap.imgW - 1 if self.idxMaj == 0 else objMap.imgH - 1
                if ptPos > rg_wgImg[0] and ptPos < rg_wgImg[1]:
                    pt = self.getPtGui2Map(ptPos, self.idxMaj)
                elif ptPos <= rg_wgImg[0]:
                    pt = 0
                elif ptPos >= rg_wgImg[1]:
                    pt = size_img - 1
                try:
                    objMap.modMajAnchor(self.idxAnc, pt)  # self.ptX
                except Exception:
                    None
            else:
                # minor axis
                if pos.y() > self.wgImg.rgY[0] and pos.y() < self.wgImg.rgY[1]:
                    ptY = self.getPtGui2Map(pos.y(), axis=1)
                elif pos.y() <= self.wgImg.rgY[0]:
                    ptY = 0
                elif pos.y() >= self.wgImg.rgY[1]:
                    ptY = objMap.imgH - 1

                ptX = self.getPtGui2Map(pos.x(), axis=0)

                if objMap.angles[1] == 0:
                    self.itc_new = ptX
                else:
                    self.itc_new = ptY - ptX * objMap.slps[1]
                try:
                    objMap.modMinAnchor(self.idxAnc, self.itc_new)  # self.ptY
                except Exception:
                    None
            self.update()

    def mouseReleaseEvent(self, event):
        pos = event.pos()
        objMap = self.grid.map
        ptX = self.getPtGui2Map(pos.x(), axis=0)
        ptY = self.getPtGui2Map(pos.y(), axis=1)
        sig = np.array(objMap.sigs[self.idx_tool])
        itc = np.array(objMap.itcs[self.idx_tool])
        if (self.idxAnc != -1 and
           event.button() == Qt.RightButton and
           self.spbTk[self.idx_tool].value() > 1):
            # remove tick
            objMap.delAnchor(self.idx_tool, self.idxAnc)
            value = self.spbTk[self.idx_tool].value() - 1
            self.switch = False
            self.spbTk[self.idx_tool].setValue(value)
            self.switch = True
        elif event.button() == Qt.LeftButton:
            if self.idx_tool == 0:
                # add tick major
                pt = ptX if self.idxMaj == 0 else ptY
                ptPress = self.ptXpress if self.idxMaj == 0 else self.ptYpress
                if ptPress == pt and abs(sig-pt).min() > sig.std() / 20:
                    objMap.addMajAnchor(pt)
                    value = self.spbTk[self.idx_tool].value() + 1
                    self.switch = False
                    self.spbTk[self.idx_tool].setValue(value)
                    self.switch = True
            else:
                # add tick minor
                # check if angle is 0 or not:
                if objMap.angles[1] == 0:
                    new_itc = ptX
                else:
                    new_itc = ptY - ptX * objMap.slps[1]
                # pass itc
                if self.ptYpress == ptY and abs(itc-new_itc).min() > itc.std() / 20:
                    objMap.addMinAnchor(new_itc)
                    value = self.spbTk[self.idx_tool].value() + 1
                    self.switch = False
                    self.spbTk[self.idx_tool].setValue(value)
                    self.switch = True

        self.update()
        self.ptX = -1
        self.ptY = -1
        self.ptXpress = -1
        self.ptYpress = -1
        self.idxAnc = -1
        self.new_itc = -1

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_A:
            self.rb_bin.setChecked(True)
        elif event.key() == Qt.Key_S:
            self.rb_rgb.setChecked(True)

    def run(self):
        self.grid.agents.setup(gmap=self.grid.map,
                               gimg=self.grid.imgs)


class WidgetAnchor(Widget_Img):
    def __init__(self, grid):
        super().__init__()
        self.grid = grid
        self.isMajV = True
        self.isMajTool = True
        self.ptMajLine = []
        self.itcs = []
        self.slp = 0
        super().make_bin_img(grid.map.imgBin)
        self.repaint()

    def paintEvent(self, event):
        painter = QPainter(self)
        super().paintImage(painter)
        pen = QPen()

        ## Major axis
        # tool switch (major)
        pen_width = 3 if self.isMajTool else 1
        pen.setWidth(pen_width)
        if not self.isMajTool:
            pen.setStyle(Qt.DotLine)
        else:
            pen.setStyle(Qt.SolidLine)
        pen.setColor(Qt.red)
        painter.setPen(pen)

        # major lines
        for pt in self.ptMajLine:
            if self.isMajV:
                painter.drawLine(int(pt), int(self.rgY[0]),
                                 int(pt), int(self.rgY[1]))
            else:
                painter.drawLine(int(self.rgX[0]), int(pt),
                                 int(self.rgX[1]), int(pt))

        ## minor axis
        for itc in self.itcs:
            if self.grid.map.angles[1] != 0:
                # if minor axis angle is not equal to 0
                x1, x2 = self.rgX 
                y1 = self.rgY[0] + itc
                y2 = y1 + (x2-x1)*self.slp
            else:
                # if minor axis angle is equal to 0
                y1, y2 = self.rgY
                x1 = self.rgX[0] + itc
                x2 = x1

            # tool switch (minor)
            pen_width = 1 if self.isMajTool else 3
            pen.setWidth(pen_width)
            if self.isMajTool:
                pen.setStyle(Qt.DotLine)
            else:
                pen.setStyle(Qt.SolidLine)
            painter.setPen(pen)

            try:
                painter.drawLine(int(x1), int(y1), int(x2), int(y2))
                pen.setWidth(3)
                pen.setStyle(Qt.SolidLine)
                painter.setPen(pen)
                if self.isMajV:
                    for x in self.ptMajLine:
                        ptX = x
                        ptY = y1 + (x - x1) * self.slp
                        if self.isInRange(ptX, ptY):
                            drawCross(ptX, ptY, painter, size=4)
                else:
                    for y in self.ptMajLine:
                        ptX = x1 + ((y - y1) / self.slp)
                        ptY = y
                        if self.isInRange(ptX, ptY):
                            drawCross(ptX, ptY, painter, size=4)
            except Exception as e:
                print(e)

        painter.end()


# class WidgetAxis(QWidget):
#     def __init__(self):
#         super().__init__()
#         self.setMouseTracking(True)
#         self.pts = []

#     def setPoints(self, pts):
#         self.pts = pts

#     def paintEvent(self, event):
#         painter = QPainter(self)
#         pen = QPen()
#         pen.setWidth(3)
#         pen.setColor(Qt.red)
#         painter.setPen(pen)
#         painter.setBrush(Qt.red)
#         # plot triangle
#         ptY = int(self.height()/2)
#         for ptX in self.pts:
#             drawTriangle(ptX, ptY, "North", painter)

#         painter.end()


def rescale(values, scaleSrc=(0, 1), scaleDst=(0, 256)):
    values = np.array(values)
    return (values-scaleSrc[0])*(scaleDst[1]-scaleDst[0])/(scaleSrc[1]-scaleSrc[0])+scaleDst[0]



# 3rd party imports
from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *

# self imports
from ..grid import *

class PnInputer(QWidget):
    """
    """
    def __init__(self, grid):
        """
        """

        super().__init__()
        self.grid = grid
        # user define
        self.gr_user = QGroupBox("User's Input (You can drag and drop files)")
        self.lo_user = QGridLayout()
        self.lb_img = QLabel()
        self.lb_map = QLabel()
        self.lb_shp = QLabel()
        self.fd_img = DnDLineEdit()
        self.fd_map = DnDLineEdit()
        self.fd_shp = DnDLineEdit()
        self.bt_img = QPushButton()
        self.bt_map = QPushButton()
        self.bt_shp = QPushButton()
        # demo
        self.gr_demo = QGroupBox("Demo")
        self.lo_demo = QVBoxLayout()
        self.lb_demo = QLabel(
            "Will use sample files to demo the program. Or go to <a href='https://poissonfish.github.io/GRID/index.html'> User Manual </a>")
        self.lb_demo.setOpenExternalLinks(True)

        # self
        self.layout = QVBoxLayout()

        self.initUI()

    def initUI(self):
        """
        """

        # USER
        ## GUI components
        self.gr_user.setCheckable(True)
        self.gr_user.setChecked(False)
        self.gr_user.clicked.connect(lambda: self.toggle(self.gr_user))
        self.lb_img.setText("Image (.tif, .jpg, .png):")
        self.lb_map.setText("Map (.csv) (OPTIONAL):")
        self.lb_shp.setText("Shape (.shp) (OPTIONAL):")
        font = self.fd_img.font()
        font.setPointSize(25)
        fm = QFontMetrics(font)
        self.fd_img.setFixedHeight(fm.height())
        self.fd_map.setFixedHeight(fm.height())
        self.fd_shp.setFixedHeight(fm.height())
        self.bt_img.setText("Browse")
        self.bt_img.clicked.connect(self.assign_PathImg)
        self.bt_map.setText("Browse")
        self.bt_map.clicked.connect(self.assign_PathMap)
        self.bt_shp.setText("Browse")
        self.bt_shp.clicked.connect(self.assign_PathShp)
        ## layout
        self.lo_user.addWidget(self.lb_img, 0, 0)
        self.lo_user.addWidget(self.fd_img, 0, 1)
        self.lo_user.addWidget(self.bt_img, 0, 2)
        self.lo_user.addWidget(self.lb_map, 1, 0)
        self.lo_user.addWidget(self.fd_map, 1, 1)
        self.lo_user.addWidget(self.bt_map, 1, 2)
        self.lo_user.addWidget(self.lb_shp, 2, 0)
        self.lo_user.addWidget(self.fd_shp, 2, 1)
        self.lo_user.addWidget(self.bt_shp, 2, 2)
        self.gr_user.setLayout(self.lo_user)

        # DEMO
        ## GUI components
        self.gr_demo.setCheckable(True)
        self.gr_demo.setChecked(True)
        self.gr_demo.clicked.connect(lambda: self.toggle(self.gr_demo))
        ## layout
        self.lo_demo.addWidget(self.lb_demo)
        self.gr_demo.setLayout(self.lo_demo)

        # LAYOUT
        self.layout.setContentsMargins(200, 50, 200, 50)
        self.layout.addWidget(self.gr_user)
        self.layout.addWidget(self.gr_demo)

        # FINALIZE
        self.setLayout(self.layout)
        self.show()

    def toggle(self, groupbox):
        """
        """

        if (groupbox.title() == "Demo"):
            self.gr_user.setChecked(not self.gr_user.isChecked())
        elif (groupbox.title() != "Demo"):
            self.gr_demo.setChecked(not self.gr_demo.isChecked())

    def assign_PathImg(self):
        """
        """

        fileter = "Images (*.tif *.jpg *.jpeg *.png)"
        path = QFileDialog().getOpenFileName(self, "", "", fileter)[0]
        self.fd_img.setText(path)

    def assign_PathMap(self):
        """
        """

        fileter = "Map (*.csv *.txt)"
        path = QFileDialog().getOpenFileName(self, "", "", fileter)[0]
        self.fd_map.setText(path)

    def assign_PathShp(self):
        """
        """

        fileter = "Shape (*.shp)"
        path = QFileDialog().getOpenFileName(self, "", "", fileter)[0]
        self.fd_shp.setText(path)

    def run(self):
        """
        """
        if self.gr_user.isChecked():
            self.grid.loadData(pathImg=self.fd_img.text(),
                               pathMap=self.fd_map.text(),
                               pathShp=self.fd_shp.text())
        else:
            self.grid.loadData()  # load demo files


class DnDLineEdit(QLineEdit):
    def __init__(self):
        super().__init__()
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(Qt.CopyAction)
            event.accept()
            text = ""
            for url in event.mimeData().urls():
                text = str(url.toLocalFile())
            self.setText(text)
        else:
            event.ignore()



import time
import matplotlib.pyplot as plt
import cv2
import numpy as np
import rasterio
import os
os.chdir("/Users/jameschen/Dropbox/photo_grid")
import grid as gd
os.chdir("/Users/jameschen/Dropbox/James Chen/GRID/Manuscript/Remote Sensing/")
os.chdir("First Revision/demo/")

# load the 1st demo file
img = gd.loadImg("demo_1.tif")


# define output function
def out_diff_size(imgIn, pts, path, size):
    img = gd.cropImg(imgIn, pts, img_W=size, img_H=size).astype(rasterio.uint8)
    bands = img.shape[2]
    with rasterio.open(path, 'w', driver="GTiff",
                       height=img.shape[0], width=img.shape[1],
                       count=bands, dtype=rasterio.uint8) as dst:
        for i in range(bands):
            dst.write(img[:, :, i], i+1)


# crop AOI for demo
pts = [[2879.464480874317, 5707.814207650273],
       [5758.928961748634, 5861.158469945355],
       [5571.508196721311, 8945.081967213115],
       [2709.0819672131147, 8808.775956284153]]

# output different file size of data from the same AOI
out_diff_size(img, pts, "size/size_2_0gb.tif", 18500)
out_diff_size(img, pts, "size/size_1_5gb.tif", 15700)
out_diff_size(img, pts, "size/size_1_0gb.tif", 13000)
out_diff_size(img, pts, "size/size_0_5gb.tif", 9150)
out_diff_size(img, pts, "size/size_0_1gb.tif", 4100)

# output different plot number of data from the same file size 500 MB
# 5*10 = 50
out_diff_size(img,
              [[2860.333586050038, 5697.028051554207],
               [2777.5966641395, 6985.360121304018],
               [4125.02653525398, 5779.764973464746],
               [4065.92873388931, 7044.457922668688]],
              "plot/plot_50.tif", 9150)

# 10*10 = 100
out_diff_size(img,
              [[2860.333586050038, 5697.028051554207],
               [4136.846095526915, 5767.945413191812],
               [3971.3722517058377, 8297.331311599697],
               [2730.3184230477636, 8226.413949962092]],
              "plot/plot_100.tif", 9150)

# 10*15 = 150
out_diff_size(img,
              [[2860.333586050038, 5708.847611827142],
               [2718.4988627748294, 8226.413949962092],
               [4621.44806671721, 8344.609552691432],
               [4763.282789992419, 5815.223654283548]],
              "plot/plot_150.tif", 9150)

# 10*20 = 200
out_diff_size(img,
              [[2848.5140257771036, 5697.028051554207],
               [2718.4988627748294, 8226.413949962092],
               [5236.06520090978, 8380.068233510236],
               [5401.539044730856, 5838.862774829417]],
              "plot/plot_200.tif", 9150)


# output benchmark results for size comparison
with open("benchmark_size.csv", "w") as f:
    f.write("size,elapse,iter\n")

dic_size = {"2_0": 2,
            "1_5": 1.5,
            "1_0": 1.0,
            "0_5": 0.5,
            "0_1": 0.1}

for i in range(100):
    for key, val in dic_size.items():
        # time the loading process
        t_cur = time.time()
        grid = gd.GRID()
        grid.loadData("size/size_%sgb.tif" % key)
        elapse = round(time.time() - t_cur, 4)
        # export results
        with open("benchmark_size.csv", "a") as f:
            f.write("%d,%.4f,%d\n" % (val, elapse, i))


# output benchmark results for plots comparison
with open("benchmark_plot.csv", "w") as f:
    f.write("plot,ep_plots,ep_seg,iter\n")

dic_plot = {"50": (50, 5, 10), 
            "100": (100, 10, 10),
            "150": (150, 10, 15),
            "200": (200, 10, 20)}

for i in range(100):
    for key, val in dic_plot.items():
        nplot, nrow, ncol = val
        grid = gd.GRID()
        grid.loadData("plot/plot_%d.tif" % nplot)
        w, h, d = grid.imgs.get("raw").shape
        grid.cropImg(pts=[[0, 0], [0, w], [h, 0], [h, w]])
        grid.binarizeImg(k=3, lsSelect=[0], valShad=0, valSmth=0)
        # locate center
        t_cur = time.time()
        grid.findPlots(nRow=nrow, nCol=ncol)
        ep_pt = round(time.time() - t_cur, 4)
        # expand boundaries
        t_cur = time.time()
        grid.cpuSeg()
        ep_sg = round(time.time() - t_cur, 4)
        # export results
        with open("benchmark_plot.csv", "a") as f:
            f.write("%d,%.4f,%.4f,%d\n" % (nplot, ep_pt, ep_sg, i))



import platform
import os


class GUser():
    """
    """
    def __init__(self):
        """
        """
        self.platform = platform.system()
        self.architecture = platform.architecture()[0]
        self.release = platform.release()
        self.machine = platform.machine()
        self.dirHome = os.path.expanduser("~")
        self.dirGrid = os.path.split(__file__)[0]

    def print_conflict_info(self):
        None

    def printInfo(self):
        print("GRID User's Info")
        print("----------------")
        print("Platform:      ",  self.platform)
        print("Architecture:  ",  self.architecture)
        print("Release:       ",  self.release)
        print("Machine:       ",  self.machine)
        print("Home Dir:      ",  self.dirHome)
        print("GRID Dir:      ",  self.dirGrid)


# print conflict message if Mac users
user = GUser()

if user.platform == "Darwin":
    print("For Mac OS users, please ignore the messages below relating to duplicated implementations. \n\
It's due to the conflict of OpenCV and PyQt, but it won't affect the performance and the results.\n")



import setuptools
setuptools.setup(name='photo_grid',
                 version='1.2.23',
                 description='A GUI for field segmentation',
                 url='https://github.com/Poissonfish/GRID',
                 python_requires='>=3.6',
                 classifiers=[
                        "Programming Language :: Python :: 3.6",
                        "Programming Language :: Python :: 3.7",
                        "Programming Language :: Python :: 3.8",
                        "Programming Language :: Python :: 3.9",
                        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                        "Operating System :: MacOS :: MacOS X",
                        "Operating System :: Microsoft :: Windows :: Windows 10"
                 ],
                 author='James Chen',
                 author_email='niche@vt.edu',
                 license='GPLv3',
                 packages=['grid', 'grid.gui'],
                 include_package_data=True,
                 install_requires=['numpy', 'pandas>=0.19.2',
                                   # data processing
                                   'h5py', 'pyshp',
                                   # math, models
                                   'sklearn', 'scipy', 'matplotlib',
                                   # image processing
                                   'image', 'opencv-python',
                                   'rasterio',
                                   # GUI
                                   'PyQt6',
                                   'qdarkstyle',
                                   # misc
                                   'tqdm'])
